package de.uulm.omi.workload.tsbs;

import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSWorkload;
import de.uulm.omi.utils.WorkloadApiProperties;
import de.uulm.omi.workload.WorkloadProcess;
import java.util.Properties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Created by Daniel Seybold on 11.05.2020.
 */
public class Runner implements Runnable {

  private final Logger LOGGER = LoggerFactory.getLogger(this.getClass());


  private final String UNIX_SHELL = "bash";
  private final String UNIX_COMMAND_PARAM = "-c";

  //TSBS constants
  private final String TSBS_RESULT_FOLDER = "TSBS/";
  private static final String TSBS_BINARY = "/opt/workloads/tsbs/bin/";

  //checking interval for TSBS commands
  private final int CHECKING_INTERVAL = 500;

  private TSBSWorkload workload;
  private String taskId;

  public Runner(TSBSWorkload workload, String taskId) {

    this.workload = workload;
    this.taskId = taskId;
  }

  @Override
  public void run() {

    //TODO: in case of failure set ProcessInstance state to error!

    //TODO: refactor to have one api call per phase and pass always the whole workload object, this will avoid the issue that checks the state in between the phases
    

    WorkloadProcess workloadProcess = WorkloadProcess.getInstance();

    Properties properties = WorkloadApiProperties.getInstance().getProperties();
    Long workloadTimestamp = System.currentTimeMillis();

    LOGGER.debug(workload.toString());

    //Phase 1: generate data
    LOGGER.debug("Starting phase 1: data generations");
    String generateDataCommand= "";
    generateDataCommand = TSBSCommandBuilder.generateDataCommand( TSBS_BINARY, workload.getDbms(), workload.getTsbsDataGeneration());
    generateDataCommand += properties.getProperty(WorkloadApiProperties.WORKLOAD_DATA_FOLDER) + workload.getDbms().toString().toLowerCase() + "-data.gz";
    LOGGER.debug("Generate data command: " + generateDataCommand);


    workloadProcess
        .startWorkload(TSBS_BINARY, UNIX_SHELL, UNIX_COMMAND_PARAM, generateDataCommand, taskId,
            workloadTimestamp);

    while (workloadProcess.isRunning()) {
      try {
        LOGGER.debug("Data generation in process...");
        Thread.sleep(CHECKING_INTERVAL);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    }

    LOGGER.debug("Phase 1: data generation finished with exit code: " +  workloadProcess.getExitCode());
    if(workloadProcess.getExitCode() != 0){
      Thread.currentThread().interrupt();
      return;
    }

    //Phase 2: query generation

    String generateQueriesCommand = "";
    generateQueriesCommand = TSBSCommandBuilder.generateQueriesCommand(TSBS_BINARY, workload.getDbms(), workload.getTsbsDataGeneration(), workload.getTsbsQueryGeneration());
    generateQueriesCommand += properties.getProperty(WorkloadApiProperties.WORKLOAD_DATA_FOLDER) + workload.getDbms().toString().toLowerCase() + "-queries.gz" ;
    LOGGER.debug(generateQueriesCommand);

    workloadProcess
        .startWorkload(TSBS_BINARY, UNIX_SHELL, UNIX_COMMAND_PARAM, generateQueriesCommand, taskId,
            workloadTimestamp);

    while (workloadProcess.isRunning()) {
      try {
        LOGGER.debug("Query generation in process...");
        Thread.sleep(CHECKING_INTERVAL);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    }
    LOGGER.debug("Phase 2: query generation finished with exit code: " + workloadProcess.getExitCode());
    if(workloadProcess.getExitCode() != 0){
      LOGGER.error("Phase 3 finished with error exit code " + workloadProcess.getExitCode());
      Thread.currentThread().interrupt();
      return;
    }

    //Phase 3: executing insert/load phase

    /*
      //clear database
      String clearDBCommand = TSBSCommandBuilder.clearDB(workload.getDbms(), workload.getTsbsLoadExecution().getDbName(),workload.getDbEndpoints());
      ProcessBuilder processClearDBBuilder = new ProcessBuilder();
      processClearDBBuilder.command(UNIX_SHELL, UNIX_COMMAND_PARAM, clearDBCommand);
      LOGGER.debug(clearDBCommand);

      Process clearDBProcess = null;
      try {
        clearDBProcess = processClearDBBuilder.start();
      } catch (IOException e) {
        e.printStackTrace();
      }
      while (clearDBProcess.isAlive()) {
        try {
          LOGGER.debug("Clear DB in process...");
          Thread.sleep(CHECKING_INTERVAL);
        } catch (InterruptedException e) {
          e.printStackTrace();
        }
      }
      LOGGER.debug("Clear DB finished with exit code: " + clearDBProcess.exitValue());

     */



    String loadCommand = "";
    loadCommand = TSBSCommandBuilder.generateLoadCommand(TSBS_BINARY, properties.getProperty(WorkloadApiProperties.WORKLOAD_DATA_FOLDER), workload.getDbms(), workload.getTsbsLoadExecution(),workload.getDbEndpoints());
    loadCommand += " &> " + properties.getProperty(WorkloadApiProperties.WORKLOAD_RESULTS_FOLDER) + TSBS_RESULT_FOLDER + taskId.trim() + "_load.txt";
    LOGGER.debug(loadCommand);
    /*
    String commandToLog = loadCommand +  System.getProperty("line.separator");
    try {
      Files.write(Paths.get(properties.getProperty(WorkloadApiProperties.WORKLOAD_RESULTS_FOLDER) + TSBS_RESULT_FOLDER + taskId.trim() + "_load.txt"),commandToLog.getBytes(StandardCharsets.UTF_8),
          StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.CREATE);
    } catch (IOException e) {
      e.printStackTrace();
      return;
    }
    */

    workloadProcess
        .startWorkload(TSBS_BINARY, UNIX_SHELL, UNIX_COMMAND_PARAM, loadCommand, taskId,
            workloadTimestamp);

    while (workloadProcess.isRunning()) {
      try {
        LOGGER.debug("Load phase in process...");
        Thread.sleep(CHECKING_INTERVAL);
      } catch (InterruptedException e) {
        LOGGER.error(e.getMessage());
        e.printStackTrace();
      }
    }
    LOGGER.debug("Phase 3: load phase finished with exit code: " + workloadProcess.getExitCode());
    if(workloadProcess.getExitCode() != 0){
      LOGGER.error("Phase 3 finished with error exit code " + workloadProcess.getExitCode());
      Thread.currentThread().interrupt();
      return;
    }



    //Phase 4: executing query phase
    //Execute Queries
    String executeQueriesCommand = TSBSCommandBuilder.executeQueriesCommand(TSBS_BINARY, properties.getProperty(WorkloadApiProperties.WORKLOAD_DATA_FOLDER), workload.getDbms(),   workload.getTsbsQueryExecution(), workload.getDbEndpoints() );
    executeQueriesCommand += " &> "  + properties.getProperty(WorkloadApiProperties.WORKLOAD_RESULTS_FOLDER) +  TSBS_RESULT_FOLDER + taskId.trim() + ".txt";
    LOGGER.debug(executeQueriesCommand);
    /*
    commandToLog = executeQueriesCommand + System.getProperty("line.separator");
    try {
      Files.write(Paths.get(properties.getProperty(WorkloadApiProperties.WORKLOAD_RESULTS_FOLDER) +  TSBS_RESULT_FOLDER + taskId.trim() + ".txt"),commandToLog.getBytes(StandardCharsets.UTF_8),
          StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.CREATE);
    } catch (IOException e) {
      e.printStackTrace();
      return;
    }
    */

    workloadProcess
        .startWorkload(TSBS_BINARY, UNIX_SHELL, UNIX_COMMAND_PARAM, executeQueriesCommand, taskId,
            workloadTimestamp);

    while (workloadProcess.isRunning()) {
      try {
        LOGGER.debug("Query phase in process...");
        Thread.sleep(CHECKING_INTERVAL);
      } catch (InterruptedException e) {
        LOGGER.error(e.getMessage());
        e.printStackTrace();
      }
    }

    if(workloadProcess.getExitCode() != 0){
      LOGGER.error("Phase 4 finished with error exit code " + workloadProcess.getExitCode());
      Thread.currentThread().interrupt();
      return;
    }

    LOGGER.debug("Phase 4: query execution finished");

  }
}
