package de.uulm.omi.workload.tsbs;

import static com.google.common.base.Preconditions.checkNotNull;

import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoint;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoints;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSDataGeneration;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSLoadExecution;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSQueryExecution;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSQueryGeneration;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSWorkload;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSBSWorkload.DbmsEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.TSDBMSDBMSProperties;
import java.util.List;
import java.util.Optional;
import java.util.StringJoiner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class TSBSCommandBuilder {

    private static final Logger LOGGER = LoggerFactory.getLogger(TSBSCommandBuilder.class);

    private static String DEFAULT_INFLUX_URLS="http://localhost:8086 ";
    private static String DEFAULT_INFLUX_BACKOFF="1s";
    private static String DEFAULT_INFLUX_CONSISTENCY="all";
    private static String DEFAULT_INFLUX_REPLICATION_FACTOR="1";


    public static String generateDataCommand(String binaryPath, TSBSWorkload.DbmsEnum dbms,  TSBSDataGeneration tsbsDataGeneration) {

        checkNotNull(tsbsDataGeneration);

        String generateDataCommand = binaryPath + "tsbs_generate_data ";
        if (tsbsDataGeneration.getUsecase() != null){
            generateDataCommand += "--use-case=\"" + tsbsDataGeneration.getUsecase().toLowerCase()+ "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --use-case parameter!");

        if (tsbsDataGeneration.getSeed() != null) {
            generateDataCommand += "--seed=" + tsbsDataGeneration.getSeed().toString().toLowerCase() + " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --seed parameter!");


        if (tsbsDataGeneration.getScale() != null) {
            generateDataCommand += "--scale=" + tsbsDataGeneration.getScale() + " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --scale parameter!");

        if (tsbsDataGeneration.getStartTime() != null) {
            generateDataCommand += "--timestamp-start=\"" + tsbsDataGeneration.getStartTime() + "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --timestamp-start parameter!");


        if (tsbsDataGeneration.getEndTime() != null) {
            generateDataCommand += "--timestamp-end=\"" + tsbsDataGeneration.getEndTime() + "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --timestamp-end parameter!");


        if (tsbsDataGeneration.getLogInterval() != null) {
            generateDataCommand += "--log-interval=\"" + tsbsDataGeneration.getLogInterval() + "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --log-interval parameter!");


        if (dbms != null) {
            generateDataCommand += "--format=\"" + dbms.toString().toLowerCase()+ "\"";
        } else throw new IllegalArgumentException("Missing or incorrect --log-interval parameter!");


        generateDataCommand += " | gzip > ";

        return generateDataCommand;

    }

    public static String generateQueriesCommand(String binaryPath, TSBSWorkload.DbmsEnum dbms, TSBSDataGeneration tsbsDataGeneration, TSBSQueryGeneration tsbsQueryGeneration) {

        checkNotNull(tsbsDataGeneration);
        checkNotNull(tsbsQueryGeneration);



        String generateQueryCommand = binaryPath + "tsbs_generate_queries ";

        if (tsbsDataGeneration.getUsecase() != null){
            generateQueryCommand += "--use-case=\"" + tsbsDataGeneration.getUsecase().toLowerCase()+ "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --use-case parameter!");

        if (tsbsDataGeneration.getSeed() != null) {
            generateQueryCommand += "--seed=" + tsbsDataGeneration.getSeed() + " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --seed parameter!");


        if (tsbsDataGeneration.getScale() != null) {
            generateQueryCommand += "--scale=" + tsbsDataGeneration.getScale() + " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --scale parameter!");

        if (tsbsDataGeneration.getStartTime() != null) {
            generateQueryCommand += "--timestamp-start=\"" + tsbsDataGeneration.getStartTime() + "\" ";
        }


        //increase last second +1
        String firstPart = tsbsDataGeneration.getEndTime().substring(0, (tsbsDataGeneration.getEndTime().length() - 2));
        String lastPart = tsbsDataGeneration.getEndTime().substring(tsbsDataGeneration.getEndTime().length() - 2,tsbsDataGeneration.getEndTime().length() - 1);
        String increasedEndTime = firstPart + (Integer.valueOf(lastPart) + 1) + "Z";
        LOGGER.debug("End time query generation: " +  increasedEndTime);

        if (increasedEndTime != null) {
            generateQueryCommand += "--timestamp-end=\"" + increasedEndTime + "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --timestamp-end parameter!");


        if(tsbsQueryGeneration.getQueries().intValue() > 0) {
            generateQueryCommand += "--queries=" + tsbsQueryGeneration.getQueries() + " ";
        }  else throw new IllegalArgumentException("Missing or incorrect --queries parameter!");


        if(tsbsQueryGeneration.getQueryType() != null) {
            generateQueryCommand += "--query-type=" + "\"" + tsbsQueryGeneration.getQueryType() + "\" ";
        }else throw new IllegalArgumentException("Missing or incorrect --query-type parameter!");


        if (dbms != null) {
            generateQueryCommand += "--format=\"" + dbms.toString().toLowerCase() + "\"";
        } else throw new IllegalArgumentException("Missing or incorrect --log-interval parameter!");


        generateQueryCommand += " | gzip > ";

        return generateQueryCommand;
    }



    public static String generateLoadCommand(String binaryPath, String dataFolder, DbmsEnum dbms,
        TSBSLoadExecution tsbsLoadExecution,
        DBEndpoints dbEndpoints) {

        checkNotNull(dbms);
        checkNotNull(tsbsLoadExecution);



        String generateLoadCommand = "cat " + dataFolder + dbms.toString().toLowerCase() + "-data.gz | gunzip | " +  binaryPath +  "tsbs_load_" + dbms.toString().toLowerCase() + " ";

        //common parameters
        if (tsbsLoadExecution.getDbName() != null){
            generateLoadCommand += "--db-name=\"" + tsbsLoadExecution.getDbName().toLowerCase()+ "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --db-name parameter!");

        if (tsbsLoadExecution.getWorkers() != null){
            generateLoadCommand += "--workers=" + tsbsLoadExecution.getWorkers() + " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --workers parameter!");

        if (tsbsLoadExecution.getReportingPeriod() != null){
            generateLoadCommand += "--reporting-period=\"" + tsbsLoadExecution.getReportingPeriod()+ "\" ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --reporting-period parameter!");

        if (tsbsLoadExecution.getBatchSize() != null){
            generateLoadCommand += "--batch-size=" + tsbsLoadExecution.getBatchSize()+ " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --batch-size parameter!");

        //DBMS specific parameters
        return getDBMSConnection(dbms, dbEndpoints, generateLoadCommand,
            tsbsLoadExecution.getDbmsPorperties(),Phase.LOAD);

    }



    public static String executeQueriesCommand(String binaryPath, String queryFolder, DbmsEnum dbms, TSBSQueryExecution tsbsQueryExecution, DBEndpoints dbEndpoints) {

        String queriesCommand = "cat " + queryFolder +  dbms.toString().toLowerCase() + "-queries.gz "   + " | gunzip | " + binaryPath + "tsbs_run_queries_" + dbms.toString().toLowerCase() + " ";

        //common parameters
        if (tsbsQueryExecution.getWorkers() != null && tsbsQueryExecution.getWorkers().intValue() > 0){
            queriesCommand += "--workers " + tsbsQueryExecution.getWorkers()+ " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --workers parameter!");

        if (tsbsQueryExecution.getBurnIn() != null){
            queriesCommand += "--burn-in " + tsbsQueryExecution.getBurnIn()+ " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --burn-in parameter!");

        if (tsbsQueryExecution.getPrintInterval() != null){
            queriesCommand += "--print-interval " + tsbsQueryExecution.getPrintInterval()+ " ";
        }
        else throw new IllegalArgumentException("Missing or incorrect --print-interval parameter!");

        if (tsbsQueryExecution.isDebug()){
            queriesCommand += "--debug 1 ";
        }

        if (tsbsQueryExecution.isPrintResponses()){
            queriesCommand += "--print-responses ";
        }



        //DBMS specific parameters
        return getDBMSConnection(dbms, dbEndpoints, queriesCommand,
            tsbsQueryExecution.getDbmsPorperties(), Phase.QUERY);

    }



    //TODO: we need to pass the current phase! load and query
    private static String getDBMSConnection(DbmsEnum dbms, DBEndpoints dbEndpoints,
        String queriesCommand, List<TSDBMSDBMSProperties> dbmsProperties,
        Phase query) {
        if(dbms.equals(DbmsEnum.INFLUX)){
            queriesCommand += influxCommand(dbmsProperties, dbEndpoints);
        }else if (dbms.equals(DbmsEnum.TIMESCALEDB)){
            queriesCommand += timescaledbCommand(dbmsProperties, dbEndpoints,query);
        }else if(dbms.equals(DbmsEnum.CASSANDRA)){
            queriesCommand += cassandraCommand(dbmsProperties,dbEndpoints,query);
        }else if(dbms.equals(DbmsEnum.MONGO)){
            queriesCommand += mongodbCommand(dbmsProperties,dbEndpoints,query);
        }

        else {
            throw new IllegalArgumentException("Currently only Influx, TimescaleDb and Cassandra are supported!");
        }

        return queriesCommand;
    }

    private static String mongodbCommand(List<TSDBMSDBMSProperties> dbmsProperties,
        DBEndpoints dbEndpoints, Phase query){

        String mongodbCommand = " ";

        StringJoiner mongosHosts = new StringJoiner(",");

        for (DBEndpoint dbEndpoint: dbEndpoints) {
            mongosHosts.add(dbEndpoint.getIpAddress()+ ":27017");
            LOGGER.warn("Custom ports for MongoDB currently not supported, using default port 27017!");

        }

        mongodbCommand += "--url " + mongosHosts;

        //TODO: check for supported but optional MongoDB commands
        //TODO: check if TSBS needs to be extended to support dedicated write and read consistency settings

        return mongodbCommand;

    }

    private static String influxCommand(
        List<TSDBMSDBMSProperties> dbmsProperties,
        DBEndpoints dbEndpoints){

        String influxCommand = " ";

        //TODO: add support for multiple Influx endpoints
        LOGGER.warn("Multiple Influx endpoints are currently not supported!");
        String influxEndpoints = "http://" + dbEndpoints.get(0).getIpAddress() + ":8086 ";
        influxCommand += "--urls " + influxEndpoints;


        //backoff
        boolean backoffCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("backoff"));
        if(backoffCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("backoff")).findAny();

            influxCommand += "--backoff " + item.get().getValue() + " ";

        }

        //consistency
        boolean consistencyCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("consistency"));
        if(consistencyCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("consistency")).findAny();

            influxCommand += "--consistency " + item.get().getValue();

        }

        //replication
        boolean replicationCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("replication-factor"));
        if(replicationCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("replication-factor")).findAny();

            influxCommand += "--replication-factor " + item.get().getValue() + " ";

        }

        return influxCommand;

    }

    private static String timescaledbCommand(List<TSDBMSDBMSProperties> dbmsProperties,
        DBEndpoints dbEndpoints, Phase query){



        String timescaledbCommand = "";

        //TSBS Timescale script requires  for load --host with multiple params, for query --hosts
        if(dbEndpoints.size() > 1){
            LOGGER.warn("Multiple TimescaleDB endpoints are currently not supported, using only the first endpoint!");
        }
        String timescaledbEndpoints = dbEndpoints.get(0).getIpAddress() + " ";

        switch (query) {
            case LOAD:
                timescaledbCommand += "--host " + timescaledbEndpoints;
                break;
            case QUERY:
                timescaledbCommand += "--hosts " + timescaledbEndpoints;
                break;
        }


        //port
        boolean portCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("port"));
        if(portCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("port")).findAny();

            timescaledbCommand += "--port " + item.get().getValue() + " ";

        }else {
            LOGGER.debug("No dedicated port provided, using default port 5432!");
            timescaledbCommand += "--port 5432 ";
        }

        //user
        boolean userCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("user"));
        if(userCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("user")).findAny();

            timescaledbCommand += "--user " + item.get().getValue() + " ";

        }else {
            throw  new IllegalArgumentException("No user parameter provided for TimescaleDB connection!");
        }

        //password
        boolean passCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("pass"));
        if(passCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("pass")).findAny();

            timescaledbCommand += "--pass " + item.get().getValue() + " ";

        }else {
            throw  new IllegalArgumentException("No pass parameter provided for TimescaleDB connection!");
        }

        //postgres params
        boolean postgresCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("postgres"));
        if(postgresCheck){
            Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                .filter(c -> c.getName().equals("postgres")).findAny();

            timescaledbCommand += "--postgres " + item.get().getValue() + " ";

        }else {
            //no SSL connections per default
            LOGGER.warn("No postgres parameters provided, falling back to default and disabling sslmode!");
            timescaledbCommand += "--postgres  sslmode=disable";
        }

        //TODO add all additional parameters supported by:
        // https://github.com/timescale/tsbs/blob/master/cmd/tsbs_load_timescaledb/main.go
        // https://github.com/timescale/tsbs/blob/master/cmd/tsbs_run_queries_timescaledb/main.go



        return  timescaledbCommand;


    }

    private static String cassandraCommand(List<TSDBMSDBMSProperties> dbmsProperties,
        DBEndpoints dbEndpoints, Phase query) {

        String cassandraCommand = "";



        //TSBS Cassandra script requires  for load --hosts with multiple params, for query --host and aggregation plan is mandatory
        switch (query) {
            case LOAD:
                StringJoiner cassandraHosts = new StringJoiner(",");

                for (DBEndpoint dbEndpoint: dbEndpoints) {
                    cassandraHosts.add(dbEndpoint.getIpAddress()+ ":9042");
                }
                cassandraCommand += "--hosts " + cassandraHosts + " ";

                //replication-factor
                boolean replicationFactorCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("replication-factor"));
                if(replicationFactorCheck){
                    Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                        .filter(c -> c.getName().equals("replication-factor")).findAny();

                    cassandraCommand += "--replication-factor " + item.get().getValue() + " ";

                }else {
                    //falling back to  default replication factor 1
                    LOGGER.warn("No replication-factor parameter provided, falling back to default 1!");
                    cassandraCommand += "--replication-factor 1 ";
                }

                //write-consistency
                boolean writeConsistencyFactorCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("consistency"));
                if(writeConsistencyFactorCheck){
                    Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                        .filter(c -> c.getName().equals("consistency")).findAny();

                    cassandraCommand += "--consistency " + item.get().getValue() + " ";

                }else {
                    //falling back to  default replication factor 1
                    LOGGER.warn("No consistency parameter provided, falling back to default ONE!");
                    cassandraCommand += "--consistency ONE ";
                }
                break;
            case QUERY:
                cassandraCommand += "--host " + dbEndpoints.get(0).getIpAddress() + ":9042 ";

                //aggregation plan
                boolean aggregationCheck = dbmsProperties.stream().anyMatch(o -> o.getName().equals("aggregation-plan"));
                if(aggregationCheck){
                    Optional<TSDBMSDBMSProperties> item = dbmsProperties.stream()
                        .filter(c -> c.getName().equals("aggregation-plan")).findAny();

                    cassandraCommand += "--aggregation-plan " + item.get().getValue() + " ";

                }else {
                    //falling back to  default aggregation-plan: server
                    LOGGER.warn("No aggregation-plan parameter provided, falling back to default server!");
                    cassandraCommand += "--aggregation-plan server ";
                }

                break;
        }


        //TODO: put here common parameters for load/query phase

        return cassandraCommand;

    }

    public static String clearDB(DbmsEnum dbms, String databaseName, DBEndpoints dbEndpoints){

        String clearCommand = "";

        if(dbms.equals(DbmsEnum.INFLUX)){
            clearCommand = " curl -X POST http://" +  dbEndpoints.get(0).getIpAddress()  + ":8086/query?q=drop%20database%20" + databaseName;
            return clearCommand;
        }


        return clearCommand;
    }
}