package de.uulm.omi.workload.ycsb;

import static com.google.common.base.Preconditions.checkNotNull;

import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoint;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DBEndpoints;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DatabaseConfig;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DatabaseConfig.DatabaseBindingEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.DatabaseConfigProperty;
import de.uulm.omi.cloudiator.dbms.evaluator.model.MeasurementConfig;
import de.uulm.omi.cloudiator.dbms.evaluator.model.MeasurementConfig.MeasurementTypeEnum;
import de.uulm.omi.cloudiator.dbms.evaluator.model.WorkloadConfig;
import de.uulm.omi.cloudiator.dbms.evaluator.model.YCSBWorkload;
import java.util.StringJoiner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Created by Daniel Seybold on 17.01.2018.
 */
public class YCSBCommandBuilder {

  private static final Logger LOGGER = LoggerFactory.getLogger(YCSBCommandBuilder.class);


  public static String  buildStartCommand(YCSBWorkload ycsbWorkload, String YcsbBinary){

    checkNotNull(ycsbWorkload);

    String startCommand = YcsbBinary;



    //add type and db binding
    startCommand += " " + ycsbWorkload.getWorkloadConfig().getWorkloadType().toString().toLowerCase();
    if (ycsbWorkload.getDatabaseConfig().getDatabaseBinding()
        .equals(DatabaseBindingEnum.CASSANDRA2)) {
      //Workaround asJaxRS Codegen of Swagger can not handle dashes in enum
      startCommand += " " + "cassandra2-cql";
    }else if(ycsbWorkload.getDatabaseConfig().getDatabaseBinding()
        .equals(DatabaseBindingEnum.COCKROACHDB)){
      startCommand += " " + "jdbc";
    }
    else{
      startCommand += " " + ycsbWorkload.getDatabaseConfig().getDatabaseBinding().toString().toLowerCase();
    }

    //add reporting status
    startCommand += " -s";

    startCommand += buildDatabaseCommand(ycsbWorkload.getDatabaseConfig(), ycsbWorkload.getDbEndpoints());

    startCommand += buildWorkloadCommand(ycsbWorkload.getWorkloadConfig());

    startCommand += buildMeasurementCommand(ycsbWorkload.getMeasurementConfig());



    return startCommand;
  }

  private static String buildDatabaseCommand(DatabaseConfig databaseConfig,
      DBEndpoints dbEndpoints){

    checkNotNull(databaseConfig);
    checkNotNull(dbEndpoints);

    String databaseCommand = "";



    //TODO: refactor in separate method
    if(databaseConfig.getDatabaseBinding().equals(DatabaseBindingEnum.COCKROACHDB)){


      databaseCommand += buildCockroachDBCommand(databaseConfig,dbEndpoints);


    }else if (databaseConfig.getDatabaseBinding().equals(DatabaseBindingEnum.MONGODB)){

      databaseCommand += buildMongoDBCommand(databaseConfig, dbEndpoints);



    }else if(databaseConfig.getDatabaseBinding().equals(DatabaseBindingEnum.ELASTICSEARCH5)){

      StringJoiner elasticsearchHosts = new StringJoiner(",");

      for (DBEndpoint dbEndpoint: dbEndpoints) {
        elasticsearchHosts.add(dbEndpoint.getIpAddress()+ ":9300");
        LOGGER.warn("Custom ports for Elasticsearch are currently not supported, using default port 9300!");

      }

      databaseCommand += " -p " + databaseConfig.getEndpointParameterName() +"=" + elasticsearchHosts.toString();

      LOGGER.debug("database command for Elasticsearch: " + databaseCommand);

      //TODO: do not call this for mongodb or postgresql, refactor to have separate builders per DBMs binding
      for(DatabaseConfigProperty databaseConfigProperty : databaseConfig.getConfigPorperties()){

        databaseCommand += " -p " + databaseConfigProperty.getName() + "=" + databaseConfigProperty.getValue();

      }

    }
    else{

      //TODO: check if ports are set and pass them to command
      StringJoiner endpoints = new StringJoiner(",");
      for(DBEndpoint endpoint : dbEndpoints){
        endpoints.add(endpoint.getIpAddress());
      }

      //Building generic DBMS connections
      databaseCommand += " -p " + databaseConfig.getEndpointParameterName() +"=" + endpoints.toString();

      databaseCommand += " -p " + databaseConfig.getTableParameterName() +"=" + databaseConfig.getTableName();

      //TODO: do not call this for mongodb or postgresql, refactor to have separate builders per DBMs binding
      for(DatabaseConfigProperty databaseConfigProperty : databaseConfig.getConfigPorperties()){

        databaseCommand += " -p " + databaseConfigProperty.getName() + "=" + databaseConfigProperty.getValue();

      }
    }




    return databaseCommand;
  }


  private static String buildCockroachDBCommand(DatabaseConfig databaseConfig,
      DBEndpoints dbEndpoints){

    String databaseCommand = "";

    StringJoiner cockroachDBConnection = new StringJoiner(";");

    //create n host permutations to load balance load equally, postgres jdbc driver only issues load against the first host and uses the remaining as fail over
    for(int i = 0; i < dbEndpoints.size(); i++){

      dbEndpoints.add(dbEndpoints.get(0));
      dbEndpoints.remove(0);

      //patched YCSB 0.12.0 JDBC binding via PostgresSQL driver
      StringJoiner cockroachEndpoint = new StringJoiner(",");
      for(DBEndpoint endpoint : dbEndpoints){
        cockroachEndpoint.add(endpoint.getIpAddress()+":26257");
      }

      cockroachDBConnection.add( "jdbc:postgresql://" + cockroachEndpoint.toString()+"/"+databaseConfig.getTableName()+"?sslmode=disable&loadBalanceHosts=true");
    }

    databaseCommand += " -p " + databaseConfig.getEndpointParameterName() + "=\"" + cockroachDBConnection.toString() + "\"";

    //add additional parameters
    for(DatabaseConfigProperty databaseConfigProperty : databaseConfig.getConfigPorperties()){

      databaseCommand += " -p " + databaseConfigProperty.getName() + "=" + databaseConfigProperty.getValue();

    }

    LOGGER.debug("database command for CockroachDB: " + databaseCommand);

    return databaseCommand;

  }

  private static String buildMongoDBCommand(DatabaseConfig databaseConfig,
      DBEndpoints dbEndpoints){


    String mongoDBConnectionUrl = "mongodb://";

    StringJoiner mongosHosts = new StringJoiner(",");

    for (DBEndpoint dbEndpoint: dbEndpoints) {
      mongosHosts.add(dbEndpoint.getIpAddress()+ ":27017");
      LOGGER.warn("Custom ports for MongoDB currently not supported, using default port 27017!");

    }

    mongoDBConnectionUrl += mongosHosts.toString();

    mongoDBConnectionUrl += "/" + databaseConfig.getTableName();

    if(!databaseConfig.getConfigPorperties().isEmpty()){
      mongoDBConnectionUrl += "?";

      StringJoiner mongoParameters = new StringJoiner("&");
      for (DatabaseConfigProperty databaseConfigProperty: databaseConfig.getConfigPorperties()) {
        mongoParameters.add(databaseConfigProperty.getName() + "=" + databaseConfigProperty.getValue());
      }

      mongoDBConnectionUrl += mongoParameters.toString();

    }

    String databaseCommand = " -p " + databaseConfig.getEndpointParameterName() + "=\"" + mongoDBConnectionUrl + "\"";

    LOGGER.debug("database command for MongoDB: " + databaseCommand);

    return databaseCommand;


  }

  private static String buildMeasurementCommand(MeasurementConfig measurementConfig){
    String measurementCommand = "";

    if(measurementConfig.getMeasurementType().equals(MeasurementTypeEnum.TIMESERIES)){

      measurementCommand += " -p measurementtype=" + measurementConfig.getMeasurementType().toString().toLowerCase();
      measurementCommand += " -p timeseries.granularity=" + measurementConfig.getInterval();

      return measurementCommand;

    }else if(measurementConfig.getMeasurementType().equals(MeasurementTypeEnum.HISTOGRAM)){
      measurementCommand += " -p measurementtype=" + measurementConfig.getMeasurementType().toString().toLowerCase();
      return measurementCommand;

    }else if (measurementConfig.getMeasurementType().equals(MeasurementTypeEnum.NONE)){
      return measurementCommand;
    }else {
      throw new IllegalStateException("Unsupported measurement type:  " + measurementConfig.getMeasurementType().toString());
    }



  }

  private static String buildWorkloadCommand(WorkloadConfig workloadConfig){
    checkNotNull(workloadConfig);

    String workloadCommand="";


    if (!workloadConfig.getWorkloadClass().isEmpty()){
      workloadCommand += " -p workload=" + workloadConfig.getWorkloadClass();
    }


    if (workloadConfig.getMaxExecutionTime() > 0){
      workloadCommand += " -p maxexecutiontime=" + workloadConfig.getMaxExecutionTime();
    }



    if (workloadConfig.getThreadCount() > 0){
      workloadCommand += " -threads " + workloadConfig.getThreadCount();
    }


    if (workloadConfig.getRecordCount() > 0){
      workloadCommand += " -p recordcount=" + workloadConfig.getRecordCount();
    }

    if (workloadConfig.getOperations() > 0){
      workloadCommand += " -p operationcount=" + workloadConfig.getOperations();
    }


    if (workloadConfig.getTargetOps() > 0){
      workloadCommand += " -target " + workloadConfig.getTargetOps();
    }



    if (workloadConfig.getFieldCount() > 0){
      workloadCommand += " -p fieldcount=" + workloadConfig.getFieldCount();
    }


    if (workloadConfig.getFieldLength() > 0){
      workloadCommand += " -p fieldlength=" + workloadConfig.getFieldLength();
    }


    if (!workloadConfig.isReadAllFileds()){
      workloadCommand += " -p readallfields=" + workloadConfig.isReadAllFileds();
    }


    if (workloadConfig.getReadModifyWriteProportion() > 0){
      workloadCommand += " -p readmodifywriteproportion=" + workloadConfig.getReadModifyWriteProportion();
    }


    if(workloadConfig.getRequestdistribution() != null && !workloadConfig.getRequestdistribution().equals("")){
      workloadCommand += " -p requestdistribution=" + workloadConfig.getRequestdistribution().toString().toLowerCase();
    }


    if(workloadConfig.getInsertOrder() != null && !workloadConfig.getInsertOrder().equals("")){
      workloadCommand += " -p insertorder=" + workloadConfig.getInsertOrder().toString().toLowerCase();
    }



    workloadCommand += " -p readproportion=" + workloadConfig.getReadProportion();

    workloadCommand += " -p updateproportion=" + workloadConfig.getUpdateProportion();

    workloadCommand += " -p insertproportion=" + workloadConfig.getInsertProportion();

    workloadCommand += " -p scanproportion=" + workloadConfig.getScanProportion();


    if (workloadConfig.getMaxScanLength() > 0){
      workloadCommand += " -p maxscanlength=" + workloadConfig.getMaxScanLength();
    }

    if(workloadConfig.getScanLengthDistribution() != null && !workloadConfig.getScanLengthDistribution().equals("")){
      workloadCommand += " -p scanlengthdistribution=" + workloadConfig.getScanLengthDistribution().toString().toLowerCase();
    }


    if (workloadConfig.getCoreWorkloadInsertionRetryLimit() > 0){
      workloadCommand += " -p core_workload_insertion_retry_limit=" + workloadConfig.getCoreWorkloadInsertionRetryLimit();
    }

    if (workloadConfig.getCoreWorkloadInsertionRetryInterval() > 0){
      workloadCommand += " -p core_workload_insertion_retry_interval=" + workloadConfig.getCoreWorkloadInsertionRetryInterval();
    }

    if (workloadConfig.getInsertStart() > 0){
      workloadCommand += " -p insertstart=" + workloadConfig.getInsertStart();
    }

    if (workloadConfig.getInsertCount() > 0){
      workloadCommand += " -p insertcount=" + workloadConfig.getInsertCount();
    }





    return workloadCommand;

  }

}
